from enum import Enum


class ActorEventType(str, Enum):
    """Possible values of actor event type."""

    #: Info about CPU usage of the actor
    CPU_INFO = 'cpuInfo'
    #: Info about resource usage of the actor
    SYSTEM_INFO = 'systemInfo'
    #: Sent when the actor is about to migrate
    MIGRATING = 'migrating'
    #: Sent when the actor should persist its state (every minute or when migrating)
    PERSIST_STATE = 'persistState'
    #: Sent when the actor is aborting
    ABORTING = 'aborting'


class ApifyEnvVars(str, Enum):
    """Possible Apify env vars."""

    ACT_ID = 'APIFY_ACT_ID'
    ACT_RUN_ID = 'APIFY_ACT_RUN_ID'
    ACTOR_BUILD_ID = 'APIFY_ACTOR_BUILD_ID'
    ACTOR_BUILD_NUMBER = 'APIFY_ACTOR_BUILD_NUMBER'
    ACTOR_EVENTS_WS_URL = 'APIFY_ACTOR_EVENTS_WS_URL'
    ACTOR_ID = 'APIFY_ACTOR_ID'
    ACTOR_RUN_ID = 'APIFY_ACTOR_RUN_ID'
    ACTOR_TASK_ID = 'APIFY_ACTOR_TASK_ID'
    API_BASE_URL = 'APIFY_API_BASE_URL'
    API_PUBLIC_BASE_URL = 'APIFY_API_PUBLIC_BASE_URL'
    CHROME_EXECUTABLE_PATH = 'APIFY_CHROME_EXECUTABLE_PATH'
    CONTAINER_PORT = 'APIFY_CONTAINER_PORT'
    CONTAINER_URL = 'APIFY_CONTAINER_URL'
    DEDICATED_CPUS = 'APIFY_DEDICATED_CPUS'
    DEFAULT_DATASET_ID = 'APIFY_DEFAULT_DATASET_ID'
    DEFAULT_KEY_VALUE_STORE_ID = 'APIFY_DEFAULT_KEY_VALUE_STORE_ID'
    DEFAULT_REQUEST_QUEUE_ID = 'APIFY_DEFAULT_REQUEST_QUEUE_ID'
    DISABLE_OUTDATED_WARNING = 'APIFY_DISABLE_OUTDATED_WARNING'
    FACT = 'APIFY_FACT'
    HEADLESS = 'APIFY_HEADLESS'
    INPUT_KEY = 'APIFY_INPUT_KEY'
    INPUT_SECRETS_PRIVATE_KEY_FILE = 'APIFY_INPUT_SECRETS_PRIVATE_KEY_FILE'
    INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE = 'APIFY_INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE'
    IS_AT_HOME = 'APIFY_IS_AT_HOME'
    LOCAL_STORAGE_DIR = 'APIFY_LOCAL_STORAGE_DIR'
    LOG_FORMAT = 'APIFY_LOG_FORMAT'
    LOG_LEVEL = 'APIFY_LOG_LEVEL'
    MEMORY_MBYTES = 'APIFY_MEMORY_MBYTES'
    META_ORIGIN = 'APIFY_META_ORIGIN'
    PROXY_HOSTNAME = 'APIFY_PROXY_HOSTNAME'
    PROXY_PASSWORD = 'APIFY_PROXY_PASSWORD'
    PROXY_PORT = 'APIFY_PROXY_PORT'
    PROXY_STATUS_URL = 'APIFY_PROXY_STATUS_URL'
    SDK_LATEST_VERSION = 'APIFY_SDK_LATEST_VERSION'
    STARTED_AT = 'APIFY_STARTED_AT'
    TIMEOUT_AT = 'APIFY_TIMEOUT_AT'
    TOKEN = 'APIFY_TOKEN'
    USER_ID = 'APIFY_USER_ID'
    WORKFLOW_KEY = 'APIFY_WORKFLOW_KEY'
    XVFB = 'APIFY_XVFB'

    # Extra ones not in consts:
    METAMORPH_AFTER_SLEEP_MILLIS = 'APIFY_METAMORPH_AFTER_SLEEP_MILLIS'
    PERSIST_STATE_INTERVAL_MILLIS = 'APIFY_PERSIST_STATE_INTERVAL_MILLIS'
    PURGE_ON_START = 'APIFY_PURGE_ON_START'


INTEGER_ENV_VARS = [
    ApifyEnvVars.CONTAINER_PORT.value,
    ApifyEnvVars.DEDICATED_CPUS.value,
    ApifyEnvVars.MEMORY_MBYTES.value,
    ApifyEnvVars.METAMORPH_AFTER_SLEEP_MILLIS.value,
    ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS.value,
    ApifyEnvVars.PROXY_PORT.value,
]

BOOL_ENV_VARS = [
    ApifyEnvVars.HEADLESS.value,
    ApifyEnvVars.IS_AT_HOME.value,
    ApifyEnvVars.PURGE_ON_START.value,
    ApifyEnvVars.XVFB.value,
]

DATETIME_ENV_VARS = [
    ApifyEnvVars.STARTED_AT.value,
    ApifyEnvVars.TIMEOUT_AT.value,
]



from datetime import datetime
from typing import Optional, cast

from ._utils import _fetch_and_parse_env_var


class Configuration:
    """TODO: docs."""

    def __init__(self) -> None:
        """TODO: docs."""
        self.actor_build_id = cast(Optional[str], _fetch_and_parse_env_var('APIFY_ACTOR_BUILD_ID'))
        self.actor_build_number = cast(Optional[str], _fetch_and_parse_env_var('APIFY_ACTOR_BUILD_NUMBER'))
        self.actor_events_ws_url = cast(Optional[str], _fetch_and_parse_env_var('APIFY_ACTOR_EVENTS_WS_URL'))
        self.actor_id = cast(Optional[str], _fetch_and_parse_env_var('APIFY_ACTOR_ID'))
        self.actor_run_id = cast(Optional[str], _fetch_and_parse_env_var('APIFY_ACTOR_RUN_ID'))
        self.actor_task_id = cast(Optional[str], _fetch_and_parse_env_var('APIFY_ACTOR_TASK_ID'))
        self.api_base_url = cast(str, _fetch_and_parse_env_var('APIFY_API_BASE_URL', 'https://api.apify.com'))
        self.api_public_base_url = cast(str, _fetch_and_parse_env_var('APIFY_API_PUBLIC_BASE_URL', 'https://api.apify.com'))
        self.chrome_executable_path = cast(Optional[str], _fetch_and_parse_env_var('APIFY_CHROME_EXECUTABLE_PATH'))
        self.container_port = cast(str, _fetch_and_parse_env_var('APIFY_CONTAINER_PORT', 4321))
        self.container_url = cast(str, _fetch_and_parse_env_var('APIFY_CONTAINER_URL', 'http://localhost:4321'))
        self.dedicated_cpus = cast(Optional[str], _fetch_and_parse_env_var('APIFY_DEDICATED_CPUS'))
        self.default_browser_path = cast(Optional[str], _fetch_and_parse_env_var('APIFY_DEFAULT_BROWSER_PATH'))
        self.default_dataset_id = cast(str, _fetch_and_parse_env_var('APIFY_DEFAULT_DATASET_ID', 'default'))
        self.default_key_value_store_id = cast(str, _fetch_and_parse_env_var('APIFY_DEFAULT_KEY_VALUE_STORE_ID', 'default'))
        self.default_request_queue_id = cast(str, _fetch_and_parse_env_var('APIFY_DEFAULT_REQUEST_QUEUE_ID', 'default'))
        self.disable_browser_sandbox = cast(Optional[str], _fetch_and_parse_env_var('APIFY_DISABLE_BROWSER_SANDBOX'))
        self.headless = cast(bool, _fetch_and_parse_env_var('APIFY_HEADLESS', True))
        self.input_key = cast(str, _fetch_and_parse_env_var('APIFY_INPUT_KEY', 'INPUT'))
        self.input_secrets_private_key_file = cast(Optional[str], _fetch_and_parse_env_var('APIFY_INPUT_SECRETS_PRIVATE_KEY_FILE'))
        self.input_secrets_private_key_passphrase = cast(Optional[str], _fetch_and_parse_env_var('APIFY_INPUT_SECRETS_PRIVATE_KEY_PASSPHRASE'))
        self.is_at_home = cast(bool, _fetch_and_parse_env_var('APIFY_IS_AT_HOME', False))
        self.memory_mbytes = cast(Optional[str], _fetch_and_parse_env_var('APIFY_MEMORY_MBYTES'))
        self.meta_origin = cast(Optional[str], _fetch_and_parse_env_var('APIFY_META_ORIGIN'))
        self.metamorph_after_sleep_millis = cast(int, _fetch_and_parse_env_var('APIFY_METAMORPH_AFTER_SLEEP_MILLIS', 300000))
        self.persist_state_interval_millis = cast(int, _fetch_and_parse_env_var('APIFY_PERSIST_STATE_INTERVAL_MILLIS', 60000))
        self.proxy_hostname = cast(str, _fetch_and_parse_env_var('APIFY_PROXY_HOSTNAME', 'proxy.apify.com'))
        self.proxy_password = cast(Optional[str], _fetch_and_parse_env_var('APIFY_PROXY_PASSWORD'))
        self.proxy_port = cast(int, _fetch_and_parse_env_var('APIFY_PROXY_PORT', 8000))
        self.proxy_status_url = cast(str, _fetch_and_parse_env_var('APIFY_PROXY_STATUS_URL', 'http://proxy.apify.com'))
        self.purge_on_start = cast(bool, _fetch_and_parse_env_var('APIFY_PURGE_ON_START', True))
        self.started_at = cast(Optional[datetime], _fetch_and_parse_env_var('APIFY_STARTED_AT'))
        self.timeout_at = cast(Optional[datetime], _fetch_and_parse_env_var('APIFY_TIMEOUT_AT'))
        self.token = cast(Optional[str], _fetch_and_parse_env_var('APIFY_TOKEN'))
        self.user_id = cast(Optional[str], _fetch_and_parse_env_var('APIFY_USER_ID'))
        self.xvfb = cast(bool, _fetch_and_parse_env_var('APIFY_XVFB', False))

        self.system_info_interval_millis = 60000
        self.max_used_cpu_ratio = 0.95



import asyncio
import inspect
import os
import sys
import time
from datetime import datetime, timezone
from typing import Any, Callable, Generic, Optional, TypeVar, Union, cast

import psutil

from apify_client import __version__ as client_version

from ._version import __version__ as sdk_version
from .consts import BOOL_ENV_VARS, DATETIME_ENV_VARS, INTEGER_ENV_VARS


def _log_system_info() -> None:
    python_version = '.'.join([str(x) for x in sys.version_info[:3]])

    print('System info:')
    print(f'    Apify SDK version: {sdk_version}')
    print(f'    Apify Client version: {client_version}')
    print(f'    OS: {sys.platform}')
    print(f'    Python version: {python_version}')


ClassPropertyType = TypeVar('ClassPropertyType')


class classproperty(Generic[ClassPropertyType]):  # noqa: N801
    """TODO: no docs."""

    def __init__(self, getter: Callable[..., ClassPropertyType]) -> None:
        """TODO: no docs."""
        self.getter = getter

    def __get__(self, _: Any, owner: Any) -> ClassPropertyType:  # noqa: U101
        """TODO: no docs."""
        return self.getter(owner)


def _fetch_and_parse_env_var(env_var_name: str, default: Any = None) -> Any:
    val = os.getenv(env_var_name)
    if not val:
        return default

    if env_var_name in BOOL_ENV_VARS:
        return _maybe_parse_bool(val)
    if env_var_name in INTEGER_ENV_VARS:
        return _maybe_parse_int(val)
    if env_var_name in DATETIME_ENV_VARS:
        return _maybe_parse_datetime(val)
    return val


def _get_cpu_usage_percent() -> float:
    return psutil.cpu_percent()


def _get_memory_usage_bytes() -> int:
    current_process = psutil.Process(os.getpid())
    mem = cast(int, current_process.memory_info().rss or 0)
    for child in current_process.children(recursive=True):
        try:
            mem += cast(int, child.memory_info().rss or 0)
        except psutil.NoSuchProcess:
            pass
    return mem


def _maybe_parse_bool(val: Optional[str]) -> bool:
    if val == 'true' or val == 'True' or val == '1':
        return True
    return False


def _maybe_parse_datetime(val: Optional[str]) -> Union[Optional[datetime], Optional[str]]:
    if not val:
        return None
    try:
        return datetime.strptime(val, '%Y-%m-%dT%H:%M:%S.%fZ').replace(tzinfo=timezone.utc)
    except ValueError:
        return val


def _maybe_parse_int(val: Optional[str]) -> Optional[int]:
    if not val:
        return None
    try:
        return int(val)
    except ValueError:
        return None


async def _run_func_at_interval_async(func: Callable, interval_secs: float) -> None:
    try:
        started_at = time.time()
        while True:
            elapsed_secs = time.time() - started_at
            sleep_for_secs = interval_secs - (elapsed_secs % interval_secs)
            await asyncio.sleep(sleep_for_secs)

            res = func()
            if inspect.isawaitable(res):
                await res
    except asyncio.CancelledError:
        pass
